/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include "listbox.h"
#include "scrollbar.h"
#include "image.h"
#include "widget_factory.h"
#include "display.h"

#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include <stdlib.h>
#include <string.h>

px_widget* listbox_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	// Initialize common widget variables

	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0,
			CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

	
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;
	widget->child_of = NULL;
	widget->wtype = LISTBOX;

	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;

	widget->use_pixmap = False;
	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");


	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);
	widget_draw_px_color(&(widget->backing_store), widget->background.background_color);


#ifdef __debug__
	printf("%s(): XCreateWindow : %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n", 
			__func__, widget, widget->win, x, y, width, height, widget->win);
	XSync(display, False);
#endif
	return widget;
}

void listbox_set_parameters(px_widget* widget)
{
	widget->default_callbacks[BUTTON_PRESSED] = listbox_default_button_press;
	widget->default_callbacks[BUTTON_RELEASED] = listbox_default_button_release;
	widget->default_callbacks[MOTION] = listbox_default_motion;
	widget->default_callbacks[MOUSE_LEAVE] = listbox_default_mouse_leave;

	widget->handle_event = listbox_handle_event;
	widget->redraw = listbox_redraw;
	widget->show = listbox_show;
	widget->destroy = listbox_destroy;


	// Initialize widget specific variables
	widget->specific.listbox.widget_list = (px_widget**)malloc(sizeof(px_widget*) * 5);
	widget->specific.listbox.event_within = NULL;
	widget->specific.listbox.event_within_index = -1;

	widget->specific.listbox.scrollbar = NULL;
	
	widget->specific.listbox.allocated = 5;
	widget->specific.listbox.no_of_children = 0;
	widget->specific.listbox.total_length_needed = 0;
	widget->specific.listbox.start_of_viewport = 0;
	widget->specific.listbox.inter_widget_space = 2;

	// Initialize widget specific callback
	widget->specific.listbox.set_inter_widget_space = listbox_inter_widget_space;
	widget->specific.listbox.set_child = listbox_set_child;
	widget->specific.listbox.remove_child = listbox_remove_child;
	widget->specific.listbox.get_child = listbox_get_child;
	widget->specific.listbox.see = listbox_see;
}

Bool listbox_handle_event(px_widget* widget, px_event* event)
{
	unsigned int count = 0;
	Bool return_value = True;
#ifdef __debug__
	printf("%s() : %p Event type = %d at %d %d\n", __func__, widget, event->type, event->xev->xbutton.x, event->xev->xbutton.y);
#endif
	if (event->type != NR_CALLBACKS && (event->type == EXPOSE || widget->get_state(widget) != DISABLED))
	{

		return_value = widget_handle_event(widget, event);

		if (event->type > 6)
		{
			for (count = 0; return_value && count < widget->specific.listbox.no_of_children; count++)
			{
				px_widget* child = widget->specific.listbox.widget_list[count];
#ifdef __debug__
				printf("%s() : %p Found widget %p at %d : hidden %d\n", __func__,widget, child, 
						count, child->hidden);
#endif
				if (!child->hidden)
					return_value = child->handle_event(child, event);
			}
		}
	}
	return return_value;
}

Bool listbox_default_button_press(px_widget* widget, px_event* event)
{
	int count;
	Bool return_value = True;
	if (listbox_within_scrollbar(widget, event))
	{
		px_widget* scrollbar = widget->specific.listbox.scrollbar;
		if (widget->specific.listbox.total_length_needed > widget->height && scrollbar != NULL)
		{
			event->x -= scrollbar->x;
			event->y -= scrollbar->y;
				
			widget->specific.listbox.event_within_index = -1;
			return_value = scrollbar->handle_event	(scrollbar, event);
			
			event->x += scrollbar->x;
			event->y += scrollbar->y;
			widget->specific.listbox.event_within = scrollbar;
		}

		return return_value;
	}

	for (count = 0; count < widget->specific.listbox.no_of_children; count++)
	{
		px_widget* child;
		
		child = widget->specific.listbox.widget_list[count];

		if (child->within(child, event))
		{
			event->x -= child->x;
			event->y -= child->y;

			widget->specific.listbox.event_within_index = count;
			return_value = child->handle_event(child, event);

			event->x += child->x;
			event->y += child->y;

			widget->specific.listbox.event_within = child;
			break;
		}
	}

	return return_value;
}

Bool listbox_default_motion(px_widget* widget, px_event* event)
{
	px_widget* in = widget->specific.listbox.event_within;
	Bool return_value = True;
	
	if (listbox_within_scrollbar(widget, event))
	{
		px_widget* scrollbar = widget->specific.listbox.scrollbar;
		if (widget->specific.listbox.total_length_needed > widget->height && scrollbar != NULL)
		{
			if (in == scrollbar)
			{
				event->x -= scrollbar->x;
				event->y -= scrollbar->y;
				
				return_value = scrollbar->handle_event	(scrollbar, event);
				event->x += scrollbar->x;
				event->y += scrollbar->y;
			}
			widget->specific.listbox.event_within = scrollbar;
			widget->specific.listbox.event_within_index = -1;
		}

		return return_value;
	}

	if (in != NULL)
	{
		if (!in->within(in, event))
		{
			event->type = MOUSE_LEAVE;

#ifdef __debug__				
			printf("%s() : %p Generating leave event on %p\n", __func__, widget, 
				widget->specific.listbox.event_within);
#endif
			in->handle_event(in, event);
			in = NULL;
			widget->specific.listbox.event_within = NULL;
			widget->specific.listbox.event_within_index = -1;
		}
		else
		{
			event->x -= in->x;
			event->y -= in->y;
			return_value = in->handle_event(in, event);
			event->x += in->x;
			event->y += in->y;
		}
	}

	if (in == NULL)
	{
		int count;
		event->type = MOUSE_ENTER;
		for (count = 0; count < widget->specific.listbox.no_of_children; count++)
		{
			px_widget* child;
			child = widget->specific.listbox.widget_list[count];

			if (child->within(child, event))
			{
				event->x -= child->x;
				event->y -= child->y;
#ifdef __debug__				
				printf("%s() : %p Generating Enter event on %p\n", __func__, widget, 
					widget->specific.listbox.event_within);
#endif
				widget->specific.listbox.event_within_index = count;
				return_value = child->handle_event(child, event);

				event->x += child->x;
				event->y += child->y;

				widget->specific.listbox.event_within = child;
				break;
			}
		}
		event->type = MOTION;
	}

	return return_value;
}

Bool listbox_default_button_release(px_widget* widget, px_event* event)
{
	px_widget* in = widget->specific.listbox.event_within;
	Bool return_value = True;

	if (listbox_within_scrollbar(widget, event))
	{
		px_widget* scrollbar = widget->specific.listbox.scrollbar;
		if (widget->specific.listbox.total_length_needed > widget->height && scrollbar != NULL)
		{
			event->x -= scrollbar->x;
			event->y -= scrollbar->y;
				
			widget->specific.listbox.event_within_index = -1;
			return_value = scrollbar->handle_event	(scrollbar, event);
			event->x += scrollbar->x;
			event->y += scrollbar->y;
			
			widget->specific.listbox.event_within = NULL;
		}

		return return_value;
	}
	
	if (in != NULL)
	{
		if (!in->within(in, event))
		{
			event->type = MOUSE_LEAVE;

#ifdef __debug__				
			printf("%s() : %p Generating leave event on %p\n", __func__, widget, 
				widget->specific.listbox.event_within);
#endif
			in->handle_event(in, event);

			event->type = BUTTON_RELEASED;
			in = NULL;
		}
		else
		{
			event->x -= in->x;
			event->y -= in->y;
			return_value = in->handle_event(in, event);
			event->x += in->x;
			event->y += in->y;
		}
	}

	if (in == NULL)
	{
		int count;
		for (count = 0; count < widget->specific.listbox.no_of_children; count++)
		{
			px_widget* child;
	
			child = widget->specific.listbox.widget_list[count];

			if (child->within(child, event))
			{
				event->x -= child->x;
				event->y -= child->y;

				widget->specific.listbox.event_within_index = count;
				return_value = child->handle_event(child, event);

				event->x += child->x;
				event->y += child->y;

				break;
			}
		}
	}
	widget->specific.listbox.event_within = NULL;
	widget->specific.listbox.event_within_index = -1;

	return return_value;
}
			
		
Bool listbox_default_mouse_leave(px_widget* widget, px_event* event)
{
	px_widget* child = widget->specific.listbox.event_within;
	Bool return_value = True;
	if (child != NULL)
	{
		return_value = child->handle_event(child, event);
		widget->specific.listbox.event_within = NULL;
		widget->specific.listbox.event_within_index = -1;
	}

	return True;
}

Bool listbox_within_scrollbar(px_widget* widget, const px_event* event)
{
	if (widget->specific.listbox.scrollbar == NULL)
		return False;

	if (event->type < 3)
	{

		if (event->x >= (widget->width - 16) && event->x <= widget->width)
		{
			return True;
		}
		return False;
	}
	return True;
}

void listbox_redraw(px_widget* widget)
{
	unsigned int count, starty, total_height, unhidden_children;
	px_widget* scrollbar = widget->specific.listbox.scrollbar;
	Display* display = display_get_Xdisplay();

#ifdef __debug__
	printf("%s() : %p : Begin\n", __func__, widget);
#endif
	if (widget->use_pixmap)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	starty = widget->specific.listbox.inter_widget_space - widget->specific.listbox.start_of_viewport;
	total_height = 0;
	unhidden_children = 0;
	for (count = 0; count < widget->specific.listbox.no_of_children; count++)
	{
		px_widget* w = widget->specific.listbox.widget_list[count];
		if(!w->hidden)
		{
			(*w->set_position)(w, 0, starty);

			starty += widget->specific.listbox.inter_widget_space + w->height;
			total_height += w->height;
			unhidden_children++;
		}
	}

	widget->specific.listbox.total_length_needed = total_height + 
		unhidden_children * widget->specific.listbox.inter_widget_space;

	if ( widget->specific.listbox.total_length_needed > widget->height)
	{
		float slider_length;
		px_pixmap back;
		GC gc, mask_gc;
		
		if (scrollbar == NULL)
		{
			scrollbar = new_widget(SCROLLBAR, widget, widget->width - 16, 0, 16, widget->height, NULL);
			scrollbar->child_of = widget;

			scrollbar->specific.scrollbar.intimate_viewport_changed = listbox_intimate_viewport_changed;
//			scrollbar->specific.scrollbar.intimate_viewport_changed = NULL;
			
			widget->specific.listbox.scrollbar = scrollbar;
		}

		slider_length = ((float)(widget->height) / (float)widget->specific.listbox.total_length_needed);
	
		scrollbar->specific.scrollbar.set_slider_length(widget->specific.listbox.scrollbar, slider_length);

		if (scrollbar->height != widget->height) {
			float slider_position;
			int ud, dd;

			dd = widget->specific.listbox.total_length_needed - widget->specific.listbox.start_of_viewport - widget->height;
			if(dd<0) {
				widget->specific.listbox.start_of_viewport += dd;
				dd = 0;
				listbox_move_children(widget);
			}
			ud = widget->specific.listbox.start_of_viewport;

			slider_position = (float)ud/(float)(ud+dd);
			scrollbar->specific.scrollbar.set_slider_position(scrollbar, slider_position);
			scrollbar->set_size(scrollbar, 16, widget->height);
		}
		if (scrollbar->x != (widget->width - 16))
			scrollbar->set_position(scrollbar, widget->width - 16, 0);

		// Draw the background
		image_initialize_pixmap(&back);
		image_create_pixmap(scrollbar->win, &back, 16, widget->height);
			
		gc = XCreateGC(display, back.pxmp, 0, NULL);
		XSetGraphicsExposures(display, gc, False);
		
		XCopyArea(display, widget->backing_store.pxmp, back.pxmp, gc, 
				widget->width - 16, 0, 16, widget->height, 0, 0);
		if (widget->backing_store.mask != None)
		{
			back.mask = XCreatePixmap(display, scrollbar->win, 10, widget->height, 1);
			mask_gc = XCreateGC(display, back.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XCopyArea(display, widget->backing_store.mask, back.mask, mask_gc, 
				widget->width - 10, 0, 10, widget->height, 0, 0);
			XFreeGC(display, mask_gc);
		}

		XFreeGC(display, gc);
		
		if (scrollbar->background.background_pixmap->w != scrollbar->width || 
				scrollbar->background.background_pixmap->h != scrollbar->height)
		{
			image_destroy(scrollbar->background.background_pixmap);
			image_initialize_pixmap(scrollbar->background.background_pixmap);
		}
		image_copy_pixmap(scrollbar->win, scrollbar->background.background_pixmap, &back);

		scrollbar->redraw(scrollbar);
		scrollbar->show(scrollbar);
	} 
	else if (scrollbar != NULL)
	{
		scrollbar->hide(scrollbar);
	}
	
	if (widget->use_pixmap == False && widget->background.background_color->type == TRANSPARENT)
	{
		GC mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
		
		XSetForeground(display, mask_gc, 1);	
		XSetGraphicsExposures(display, mask_gc, False);
		for (count = 0; count < widget->specific.listbox.no_of_children; count++)
		{
			px_widget* w = widget->specific.listbox.widget_list[count];
			
			XCopyArea(display, w->backing_store.mask,
					widget->backing_store.mask, mask_gc, 0, 0, 
					w->backing_store.w, w->backing_store.h,
					w->x, w->y);
		}

		if (widget->specific.listbox.total_length_needed > widget->height && scrollbar != NULL)
		{
			XCopyArea(display, scrollbar->backing_store.mask,
					widget->backing_store.mask, mask_gc, 0, 0, 
					scrollbar->backing_store.w, scrollbar->backing_store.h,
					scrollbar->x, scrollbar->y);
		}
		XFreeGC(display, mask_gc);
	}
	widget_draw_border(widget);
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);
	
#ifdef __debug__
	XSync(display, False);
	printf("%s(): Redrew widget %p \n", __func__, widget);
#endif
}

void listbox_show(px_widget* widget)
{
	unsigned int count;
	Display* display = display_get_Xdisplay();
	XSetWindowBackgroundPixmap(display, widget->win, widget->backing_store.pxmp);

	if (widget->backing_store.mask != None)
		XShapeCombineMask (display, widget->win, ShapeBounding, 0, 0, 
			widget->backing_store.mask, ShapeSet);

	XMapWindow(display, widget->win);
	for (count = 0; count < widget->specific.listbox.no_of_children; count++)
	{
		px_widget* w = widget->specific.listbox.widget_list[count];
		
		if (!w->hidden && (w->y + w->height) > 0 && w->y <= widget->height)
		{
			(*w->show)(w);
		}
	}

	if (widget->specific.listbox.total_length_needed > widget->height)
	{
		widget->specific.listbox.scrollbar->show(widget->specific.listbox.scrollbar);
	}
	widget->hidden = False;

}

void listbox_destroy(px_widget* widget)
{
	int count;

	for (count = 0; count < widget->specific.listbox.no_of_children; count++)
	{
		(*widget->specific.listbox.widget_list[count]->destroy)
			(widget->specific.listbox.widget_list[count]);
		free(widget->specific.listbox.widget_list[count]);
	}
	free(widget->specific.listbox.widget_list);

	if (widget->specific.listbox.scrollbar != NULL)
	{
		widget->specific.listbox.scrollbar->destroy(widget->specific.listbox.scrollbar);
		free(widget->specific.listbox.scrollbar);
	}

	widget_destroy(widget);
}

void listbox_inter_widget_space(px_widget* widget, int spc)
{
	widget->specific.listbox.inter_widget_space = spc;
}

void listbox_set_child(px_widget* widget, px_widget* child, unsigned int index)
{
	unsigned int count;
	Display* display = display_get_Xdisplay();
	
	if (index > widget->specific.listbox.no_of_children)
		index = widget->specific.listbox.no_of_children;
	
	if (widget->specific.listbox.allocated == widget->specific.listbox.no_of_children)
	{
		px_widget** new_list = (px_widget**)malloc(sizeof(px_widget*) * 
				(widget->specific.listbox.allocated + 10));
			
		for (count = 0; count < widget->specific.listbox.no_of_children; count++)
			new_list[count] = widget->specific.listbox.widget_list[count];

		free(widget->specific.listbox.widget_list);
		widget->specific.listbox.widget_list = new_list;
		widget->specific.listbox.allocated += 10;
	}
	
	if (index < widget->specific.listbox.no_of_children)
	{
		(*child->set_position)(child, 2, widget->specific.listbox.widget_list[index]->y);
		for (count = widget->specific.listbox.no_of_children; count > index; count--)
		{
			widget->specific.listbox.widget_list[count] = widget->specific.listbox.widget_list[count-1];
		}

	}

	widget->specific.listbox.widget_list[index] = child;
	widget->specific.listbox.no_of_children++;
	XLowerWindow(display, child->win);
}

void listbox_remove_child(px_widget* widget, unsigned int index)
{
	
	if (index < widget->specific.listbox.no_of_children)
	{
		unsigned int count;

		if (index < widget->specific.listbox.no_of_children - 1)
		{
			for (count = index; count < widget->specific.listbox.no_of_children - 1; count++)
			{
				widget->specific.listbox.widget_list[count] = widget->specific.listbox.widget_list[count+1];
			}
		}
		
		widget->specific.listbox.no_of_children--;
	}
}

px_widget* listbox_get_child(px_widget* widget, unsigned int index)
{
	if (index < widget->specific.listbox.no_of_children)
	{
		return widget->specific.listbox.widget_list[index];
	}

	return NULL;
}

void listbox_move_children(px_widget* widget)
{
	px_listbox* listbox = &(widget->specific.listbox);
	unsigned int count;
	int starty = 0;
	
	starty = listbox->inter_widget_space - listbox->start_of_viewport;
	for (count = 0; count < listbox->no_of_children; count++)
	{
		px_widget* w = listbox->widget_list[count];
		if(!w->hidden)
		{
			w->set_position(w, 0, starty);

			starty += listbox->inter_widget_space + w->height;
		}
	}

	if (widget->use_pixmap == False && widget->background.background_color->type == TRANSPARENT)
	{
		Display* display = display_get_Xdisplay();
		GC mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
		px_widget* scrollbar = listbox->scrollbar;
		
		XSetForeground(display, mask_gc, 1);	
		XSetGraphicsExposures(display, mask_gc, False);
		for (count = 0; count < listbox->no_of_children; count++)
		{
			px_widget* w = listbox->widget_list[count];
			
			XCopyArea(display, w->backing_store.mask,
					widget->backing_store.mask, mask_gc, 0, 0, 
					w->backing_store.w, w->backing_store.h,
					w->x, w->y);
		}

		if (scrollbar != NULL && !scrollbar->hidden)
		{
			XCopyArea(display, scrollbar->backing_store.mask,
					widget->backing_store.mask, mask_gc, 0, 0, 
					scrollbar->backing_store.w,
					scrollbar->backing_store.h,
					scrollbar->x,
					scrollbar->y);
		}
		XFreeGC(display, mask_gc);
		XClearWindow(display, widget->win);
		widget_set_event_mask(widget, widget->emask);
	}
}

void listbox_see(px_widget* widget, int item)
{
	px_listbox* listbox  = &(widget->specific.listbox);

	if (item < listbox->no_of_children && !listbox->widget_list[item]->hidden &&
			((listbox->total_length_needed < widget->height && listbox->start_of_viewport > 0) ||
			 	listbox->total_length_needed > widget->height))
	{
		unsigned int count;
		int starty = 0;
		float slider_position;
		int ud, dd;

		for (count = 0; count < item; count++) {
			px_widget* w = listbox->widget_list[count];

			if (!w->hidden) {
				starty += w->height;
			}
		}
		starty += listbox->inter_widget_space * (item + 1);		

		if(starty < listbox->start_of_viewport) {
			listbox->start_of_viewport = starty;
			listbox_move_children(widget);
		} else if((starty+listbox->widget_list[item]->height) > (listbox->start_of_viewport+widget->height-1)) {
			if(listbox->widget_list[item]->height > widget->height) {
				listbox->start_of_viewport = starty;
			} else {
				listbox->start_of_viewport = starty + listbox->widget_list[item]->height - widget->height;
			}
			listbox_move_children(widget);
		}

		dd = widget->specific.listbox.total_length_needed - widget->specific.listbox.start_of_viewport - widget->height;
		if(dd<0) {
			//widget->specific.listbox.start_of_viewport += dd;
			widget->specific.listbox.start_of_viewport = 0;
			dd = 0;
			listbox_move_children(widget);
		}
		ud = widget->specific.listbox.start_of_viewport;

		slider_position = (float)ud/(float)(ud+dd);
	
		if(listbox->scrollbar != NULL) {
			if (listbox->total_length_needed > widget->height) {
				listbox->scrollbar->specific.scrollbar.set_slider_position(listbox->scrollbar, slider_position);
				listbox->scrollbar->redraw(listbox->scrollbar);
				listbox->scrollbar->show(listbox->scrollbar);
			} else {
				listbox->scrollbar->hide(listbox->scrollbar);
			}
		}
	}
}
				
void listbox_intimate_viewport_changed(px_widget* widget, float slider_start)
{
	widget->specific.listbox.start_of_viewport = (widget->specific.listbox.total_length_needed-widget->height) * slider_start;
	listbox_move_children(widget);
}
